@@ -7,8 +7,9 @@ use Scalar::Util qw(blessed);
 use HTTP::OAI;
 use Carp;
 use Catmandu::Error;
+use URI;
 
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 
 with 'Catmandu::Importer';
 
@@ -23,8 +24,13 @@ has resumptionToken => (is => 'ro');
 has oai     => (is => 'ro', lazy => 1, builder => 1);
 has dry     => (is => 'ro');
 has listIdentifiers => (is => 'ro');
+has listSets => (is => 'ro');
 has max_retries => ( is => 'ro', default => sub { 0 } );
 has _retried => ( is => 'rw', default => sub { 0; } );
+has _xml_handlers => ( is => 'ro', default => sub { +{} } );
+has realm => ( is => 'ro', predicate => 1 );
+has username => ( is => 'ro', predicate => 1 );
+has password => ( is => 'ro', predicate => 1 );
 
 sub _build_handler {
     my ($self) = @_;
@@ -74,11 +80,71 @@ sub _coerce_xslt {
 
 sub _build_oai {
     my ($self) = @_;
-    my $agent = HTTP::OAI::Harvester->new(baseURL => $self->url, resume => 0);
+    my $agent = HTTP::OAI::Harvester->new(baseURL => $self->url, resume => 0, keep_alive => 1);
+    if( $self->has_username && $self->has_password ) {
+
+        my $uri = URI->new( $self->url );
+        my @credentials = (
+            $uri->host_port,
+            $self->realm || "",
+            $self->username,
+            $self->password
+        );
+        $agent->credentials( @credentials );
+    }
     $agent->env_proxy;
     $agent;
 }
 
+sub _xml_handler_for_node {
+    my ( $self, $node ) = @_;
+    my $ns = $node->namespaceURI();
+
+    my $type;
+
+    if( $ns eq "http://www.openarchives.org/OAI/2.0/oai_dc/" ){
+
+        $type = "oai_dc";
+
+    }
+    elsif( $ns eq "http://www.loc.gov/MARC21/slim" ){
+
+        $type = "marcxml";
+
+    }
+    elsif( $ns eq "http://www.loc.gov/mods/v3" ){
+
+        $type = "mods";
+
+    }
+    else{
+
+        $type = "struct";
+
+    }
+
+    $self->_xml_handlers()->{$type} ||= Catmandu::Util::require_package( "Catmandu::Importer::OAI::Parser::$type" )->new();
+}
+
+sub _map_set {
+    my ($self, $rec) = @_;
+
+    +{
+        _id => $rec->setSpec(),
+        setSpec => $rec->setSpec(),
+        setName => $rec->setName(),
+        setDescription => [ map {
+
+            #root: 'setDescription'
+            my @root = $_->dom()->childNodes();
+            #child: oai_dc, marcxml, mods..
+            my @children = $root[0]->childNodes();
+            $self->_xml_handler_for_node( $children[0] )->parse( $children[0] );
+
+        } $rec->setDescription() ]
+    };
+}
+
 sub _map_record {
     my ($self, $rec) = @_;
 
@@ -108,8 +174,8 @@ sub _map_record {
     $data;
 }
 
-sub _args {
-    my ($self) = @_;
+sub _args_for_records {
+    my $self = $_[0];
 
     my %args = (
         metadataPrefix => $self->metadataPrefix,
@@ -125,6 +191,31 @@ sub _args {
     return %args;
 }
 
+sub _args {
+    my $self = $_[0];
+
+    my %args;
+
+    if( $self->listSets() ){
+
+    }
+    else{
+        %args = $self->_args_for_records();
+    }
+
+    %args;
+}
+
+sub _verb {
+    my $self = $_[0];
+
+    $self->listIdentifiers ?
+        'ListIdentifiers' :
+        $self->listSets ?
+            'ListSets' :
+            'ListRecords';
+}
+
 sub handle_record {
     my ($self, $dom) = @_;
     return unless $dom;
@@ -142,17 +233,58 @@ sub dry_run {
         return if $called;
         $called = 1;
         # TODO: make sure that HTTP::OAI does not change this internal method
-        return {
+        return +{
             url => $self->oai->_buildurl(
-                $self->_args,
-                verb => ($self->listIdentifiers ? 'ListIdentifiers' : 'ListRecords')
+                $self->_args(),
+                verb => $self->_verb()
             )
         };
     };
 }
 
-sub oai_run {
-    my ($self) = @_;
+sub _retry {
+    my ( $self, $sub ) = @_;
+
+    $self->_retried( 0 );
+
+    my $res;
+
+    while ( 1 ) {
+
+        $res = $sub->();
+
+        if ($res->is_error) {
+
+            my $max_retries = $self->max_retries();
+            my $_retried = $self->_retried();
+
+            if ( $max_retries > 0 && $_retried < $max_retries  ){
+
+                $_retried++;
+
+                #exponential backoff:  [0 .. 2^c [
+                my $n_seconds = int( 2**$_retried );
+                $self->log->error("failed, retrying after $n_seconds");
+                sleep $n_seconds;
+                $self->_retried( $_retried );
+                next;
+            }
+            else{
+                my $err_msg = $self->url . " : " . $res->message." (stopped after ".$self->_retried()." retries)";
+                $self->log->error( $err_msg );
+                Catmandu::Error->throw( $err_msg );
+            }
+
+        }
+
+        last;
+    }
+
+    $res;
+}
+
+sub _list_records {
+    my $self = $_[0];
     sub {
         state $stack = [];
         state $resumptionToken = $self->resumptionToken;
@@ -172,45 +304,16 @@ sub oai_run {
                 %args = (verb => $verb , resumptionToken => $resumptionToken);
             }
 
-            $self->_retried( 0 );
+            my $sub = $self->listIdentifiers() ?
+                sub { $self->oai->ListIdentifiers( %args , onRecord => $fill_stack ); } :
+                sub { $self->oai->ListRecords( %args , onRecord => $fill_stack ); };
 
-            while(1){
-
-                my $res = $self->listIdentifiers
-                    ? $self->oai->ListIdentifiers( %args , onRecord => $fill_stack )
-                    : $self->oai->ListRecords( %args , onRecord => $fill_stack );
-
-                if ($res->is_error) {
-
-                    my $max_retries = $self->max_retries();
-                    my $_retried = $self->_retried();
-
-                    if ( $max_retries > 0 && $_retried < $max_retries  ){
-
-                        $_retried++;
-
-                        #exponential backoff:  [0 .. 2^c [
-                        my $n_seconds = int( 2**$_retried );
-                        $self->log->error("failed, retrying after $n_seconds");
-                        sleep $n_seconds;
-                        $self->_retried( $_retried );
-                        next;
-                    }
-                    else{
-                        my $token = $resumptionToken // '';
-                        my $err_msg = $self->url . " : $token : " . $res->message." (stopped after ".$self->_retried()." retries)";
-                        $self->log->error( $err_msg );
-                        Catmandu::Error->throw( $err_msg );
-                    }
-                }
-
-                if (defined $res->resumptionToken) {
-                    $resumptionToken = $res->resumptionToken->resumptionToken;
-                }
-                else {
-                    $resumptionToken = undef;
-                }
-                last;
+            my $res = $self->_retry( $sub );
+            if (defined $res->resumptionToken) {
+                $resumptionToken = $res->resumptionToken->resumptionToken;
+            }
+            else {
+                $resumptionToken = undef;
             }
 
             unless (defined $resumptionToken && length $resumptionToken) {
@@ -221,7 +324,8 @@ sub oai_run {
         if (my $rec = shift @$stack) {
             if ($rec->isa('HTTP::OAI::Record')) {
                 return $self->_map_record($rec);
-            } else {
+            }
+            else {
                 return {
                     _id => $rec->identifier,
                     _datestamp  => $rec->datestamp,
@@ -234,8 +338,51 @@ sub oai_run {
     };
 }
 
+sub _list_sets {
+    my $self = $_[0];
+    sub {
+        state $stack = [];
+        state $done  = 0;
+
+        my $fill_stack = sub {
+            push @$stack , shift;
+        };
+
+        if (@$stack <= 1 && $done == 0) {
+            my %args = $self->_args;
+
+            my $sub = sub { $self->oai->ListSets( onRecord => $fill_stack ); };
+
+            my $res = $self->_retry( $sub );
+            $done = 1;
+        }
+
+        if (my $rec = shift @$stack) {
+            return $self->_map_set($rec);
+        }
+
+        return undef;
+    };
+}
+
+sub oai_run {
+    my ($self) = @_;
+
+    $self->listSets() ?
+        $self->_list_sets() :
+        $self->_list_records();
+
+}
+
 sub generator {
     my ($self) = @_;
+
+    if( $self->listIdentifiers() && $self->listSets() ){
+
+        croak "options listIdentifiers and listSets cannot be set both";
+
+    }
+
     return $self->dry ? $self->dry_run : $self->oai_run;
 }
 
@@ -316,6 +463,10 @@ for datestamp-based selective harvesting.
 
 Harvest identifiers instead of full records.
 
+=item listSets
+
+Harvest sets instead of records
+
 =item resumptionToken
 
 An optional resumptionToken to start harvesting from.
@@ -336,16 +487,34 @@ Set to '0' by default.
 
 Internally the exponential backoff algorithm is used
 for this. This means that after every failed request the importer
-sleeps for 2^collision seconds. The maximum ammount of time after which
-the importer stops can be calculated with:
+will choose a random number between 0 and 2^collision (excluded),
+and wait that number of seconds. So the actual ammount of time before
+the importer stops can differ:
 
- max_retries = 1 -> max = 2 seconds
- max_retries = 2 -> max = 2 + 4 = 6 seconds
- max_retries = 3 -> max = 2 + 4 + 8 = 14 seconds
- .
- .
- .
- max_retries = n -> max = 2^(n + 1) - 2 seconds
+ first retry:
+    wait [ 0..2^1 [ seconds
+ second retry:
+    wait [ 0..2^2 [ seconds
+ third retry:
+    wait [ 0..2^3 [ seconds
+
+ ..
+
+=item realm
+
+An optional realm value. This value is used when the importer harvests from a
+repository which is secured with basic authentication through Integrated Windows
+Authentication (NTLM or Kerberos).
+
+=item username
+
+An optional username value. This value is used when the importer harvests from a
+repository which is secured with basic authentication.
+
+=item password
+
+An optional password value. This value is used when the importer harvests from a
+repository which is secured with basic authentication.
 
 =back
 
